import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Ticket',
            fields=[
                (
                    'id',
                    models.BigAutoField(
                        primary_key=True,
                        serialize=False,
                        unique=True,
                        verbose_name='ID',
                    ),
                ),
                (
                    'created_at',
                    models.DateTimeField(
                        auto_now_add=True, verbose_name='Создан в'
                    ),
                ),
                (
                    'title',
                    models.CharField(max_length=256, verbose_name='Заголовок'),
                ),
                (
                    'resolved',
                    models.BooleanField(default=False, verbose_name='Решен'),
                ),
                (
                    'support_manager',
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name='tickets',
                        to=settings.AUTH_USER_MODEL,
                        verbose_name='Менеджер поддержки',
                    ),
                ),
                (
                    'user',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='tickets',
                        to=settings.SUPPORT_TELEGRAM_USER_MODEL,
                        verbose_name='Пользователь',
                    ),
                ),
            ],
            options={
                'verbose_name': 'Тикет',
                'verbose_name_plural': 'Тикеты',
            },
        ),
        migrations.CreateModel(
            name='Message',
            fields=[
                (
                    'id',
                    models.BigAutoField(
                        primary_key=True,
                        serialize=False,
                        unique=True,
                        verbose_name='ID',
                    ),
                ),
                (
                    'created_at',
                    models.DateTimeField(
                        auto_now_add=True, verbose_name='Создано в'
                    ),
                ),
                (
                    'sender',
                    models.CharField(
                        choices=[
                            ('user', 'Пользователь'),
                            ('supp', 'Менеджер поддержки'),
                        ],
                        default='user',
                        max_length=4,
                        verbose_name='Отправитель',
                    ),
                ),
                (
                    'text',
                    models.CharField(max_length=4096, verbose_name='Текст'),
                ),
                (
                    'viewed',
                    models.BooleanField(
                        default=False, verbose_name='Просмотрено'
                    ),
                ),
                (
                    'ticket',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='messages',
                        to='support.ticket',
                        verbose_name='Тикет',
                    ),
                ),
            ],
            options={
                'verbose_name': 'Сообщение',
                'verbose_name_plural': 'Сообщения',
            },
        ),
    ]



from django_filters import rest_framework as filters

from . import models


class Ticket(filters.FilterSet):
    user_id = filters.NumberFilter(
        field_name='user',
    )
    manager = filters.NumberFilter(
        field_name='support_manager',
    )
    manager__isnull = filters.BooleanFilter(
        field_name='support_manager',
        lookup_expr='isnull',
    )

    class Meta:
        model = models.Ticket
        fields = [
            'user_id',
            'resolved',
            'manager',
        ]



from django.urls import path

from . import views

urlpatterns = [
    path('manager/', views.Support.as_view(), name='manager'),
    path('tickets/', views.TicketList.as_view(), name='tickets'),
    path('tickets/<int:ticket_id>/', views.Ticket.as_view(), name='ticket'),
    path(
        'tickets/<int:ticket_id>/messages/',
        views.MessageList.as_view(),
        name='messages',
    ),
]



from channels.db import database_sync_to_async
from channels.generic.websocket import AsyncJsonWebsocketConsumer
from channels.layers import BaseChannelLayer, get_channel_layer
from django.contrib.auth import get_user_model

from . import models, serializers


class BaseChatConsumer(AsyncJsonWebsocketConsumer):
    unassigned_tickets_group_name = 'unassigned_tickets'

    @staticmethod
    def get_ticket_group_name(ticket_id: int) -> str:
        return f'ticket_{ticket_id}'

    @classmethod
    def get_channel_layer(cls) -> BaseChannelLayer:
        return get_channel_layer(cls.channel_layer_alias)

    @property
    def _channel_layer(self) -> BaseChannelLayer:
        return self.channel_layer

    async def add_to_group(self, group_name: str):
        await self._channel_layer.group_add(group_name, self.channel_name)

    async def remove_from_group(self, group_name: str):
        await self._channel_layer.group_discard(group_name, self.channel_name)

    async def add_ticket_to_group(self, ticket_id):
        group_name = self.get_ticket_group_name(ticket_id)
        await self.add_to_group(group_name)
        self.get_ticket_group_names_set.add(group_name)

    async def remove_ticket_from_group(self, ticket_id):
        group_name = self.get_ticket_group_name(ticket_id)
        await self.remove_from_group(group_name)
        self.get_ticket_group_names_set.discard(group_name)

    @database_sync_to_async
    def get_ticket_ids(self):
        return list(
            models.Ticket.objects.filter(
                support_manager=self.manager
            ).values_list('id', flat=True)
        )

    async def connect(self):
        User = get_user_model()  # noqa: N806
        self.get_ticket_group_names_set = set()
        try:
            self.manager = await User.objects.aget(pk=self.scope['user'].pk)
        except (User.DoesNotExist, AttributeError, KeyError):
            return await self.close()
        await self.add_to_group(self.unassigned_tickets_group_name)
        for ticket_id in await self.get_ticket_ids():
            await self.add_ticket_to_group(ticket_id)
        await self.accept()

    async def disconnect(self, code):  # noqa: ARG002
        await self.remove_from_group(self.unassigned_tickets_group_name)
        for ticket_id in self.get_ticket_group_names_set:
            await self.remove_ticket_from_group(ticket_id)


class ChatConsumerSerializerMixin:
    @database_sync_to_async
    def serialize_message(self, message: models.Message) -> dict:
        return serializers.Message(message).data

    @database_sync_to_async
    def serialize_messages(self, ticket_id: int) -> list[dict]:
        return serializers.Message(
            models.Message.objects.filter(
                ticket__id=ticket_id,
            ).order_by('-created_at'),
            many=True,
        ).data


class ChatConsumerSender(ChatConsumerSerializerMixin, BaseChatConsumer):
    async def ticket_created(self, event: dict):
        event['ticket'] = serializers.Ticket(event['ticket']).data
        await self.send_json(event)

    async def ticket_assigned(self, event: dict):
        if self.manager.pk == event['support_manager']:
            await self.add_ticket_to_group(event['id'])
        await self.send_json(event)

    async def ticket_message_new(self, event: dict):
        event['message'] = await self.serialize_message(event['message'])
        await self.send_json(event)


class ChatConsumerMessages:
    @staticmethod
    def ticket_assigned(ticket: models.Ticket, manager):
        return (
            BaseChatConsumer.unassigned_tickets_group_name,
            {
                'type': 'ticket.assigned',
                'id': ticket.id,
                'support_manager': manager.pk,
            },
        )

    @staticmethod
    def ticket_created(ticket: models.Ticket):
        return (
            BaseChatConsumer.unassigned_tickets_group_name,
            {
                'type': 'ticket.created',
                'ticket': ticket,
            },
        )

    @staticmethod
    def ticket_message_new(message: models.Message):
        return (
            ChatConsumer.get_ticket_group_name(message.ticket.id),
            {
                'type': 'ticket.message.new',
                'message': message,
            },
        )


class ChatConsumerReceiver(BaseChatConsumer):
    async def _receive_ticket_assign(self, *, ticket_id: int):
        try:
            ticket = await models.Ticket.objects.aget(
                id=ticket_id,
            )
        except models.Ticket.DoesNotExist:
            return None, None
        ticket.support_manager = self.manager
        await ticket.asave()
        return ChatConsumerMessages.ticket_assigned(ticket, self.manager)

    async def _receive_ticket_message_new(self, *, ticket_id: int, text: str):
        try:
            ticket = await models.Ticket.objects.aget(
                id=ticket_id,
            )
        except models.Ticket.DoesNotExist:
            return None, None
        await models.Message.objects.acreate(
            ticket=ticket,
            sender=models.Message.Sender.SUPPORT_MANAGER,
            text=text,
            viewed=True,
        )
        return None, None

    async def none_handler(self, *_, **__):
        return None, None


class ChatConsumer(ChatConsumerReceiver, ChatConsumerSender):
    async def get_group_and_message(
        self, content: dict
    ) -> tuple[str, dict] | tuple[None, None]:
        handler = {
            'ticket.assign': self._receive_ticket_assign,
            'ticket.message.new': self._receive_ticket_message_new,
        }.get(
            content.pop('type', None),
            self.none_handler,
        )
        return await handler(**content)

    async def receive_json(self, content: dict, **_):
        group, message = await self.get_group_and_message(content)
        if all([group, message]):
            await self._channel_layer.group_send(
                group,
                message,
            )



from . import models
from .consumers import ChatConsumer, ChatConsumerMessages
from asgiref.sync import async_to_sync


def post_save_ticket(sender, instance: models.Ticket, created: bool, **kwargs):  # noqa: ARG001
    if created:
        async_to_sync(ChatConsumer.get_channel_layer().group_send)(
            *ChatConsumerMessages.ticket_created(instance)
        )
        if instance.support_manager:
            async_to_sync(ChatConsumer.get_channel_layer().group_send)(
                *ChatConsumerMessages.ticket_assigned(
                    instance,
                    instance.support_manager,
                )
            )


def post_save_message(
    sender,  # noqa: ARG001
    instance: models.Message,
    created: bool,
    **kwargs,  # noqa: ARG001
):
    if created:
        async_to_sync(ChatConsumer.get_channel_layer().group_send)(
            *ChatConsumerMessages.ticket_message_new(instance)
        )



from django.apps import apps
from django.conf import settings
from django.core.checks import Error, register


@register()
def check_telegram_user_model(app_configs, **kwargs):  # noqa: ARG001
    errors = []
    model_path = getattr(settings, 'SUPPORT_TELEGRAM_USER_MODEL', None)

    if not model_path:
        errors.append(
            Error(
                'SUPPORT_TELEGRAM_USER_MODEL не указан в settings.py',
                id='support.E001',
            )
        )
        return errors

    try:
        model = apps.get_model(model_path, require_ready=False)
    except LookupError:
        errors.append(
            Error(
                f"SUPPORT_TELEGRAM_USER_MODEL '{model_path}' не найден",
                id='support.E002',
            )
        )
        return errors

    if not hasattr(model, 'telegram_id'):
        errors.append(
            Error(
                f"SUPPORT_TELEGRAM_USER_MODEL '{model_path}' должен содержать "
                "поле 'telegram_id'",
                id='support.E003',
            )
        )

    return errors



from django.conf import settings
from django.db import models


def get_telegram_user_model():
    """
    Возвращает модель пользователя Telegram, указанную в
    `settings.SUPPORT_TELEGRAM_USER_MODEL`.
    """
    from django.apps import apps

    return apps.get_model(
        settings.SUPPORT_TELEGRAM_USER_MODEL, require_ready=False
    )


class Ticket(models.Model):
    id = models.BigAutoField(
        verbose_name='ID',
        primary_key=True,
        unique=True,
    )
    created_at = models.DateTimeField(
        verbose_name='Создан в',
        auto_now_add=True,
    )
    user = models.ForeignKey(
        verbose_name='Пользователь',
        to=settings.SUPPORT_TELEGRAM_USER_MODEL,
        on_delete=models.CASCADE,
        related_name='tickets',
    )
    title = models.CharField(
        verbose_name='Заголовок',
        max_length=256,
    )
    support_manager = models.ForeignKey(
        verbose_name='Менеджер поддержки',
        to=settings.AUTH_USER_MODEL,
        on_delete=models.SET_NULL,
        related_name='tickets',
        null=True,
        blank=True,
    )
    resolved = models.BooleanField(
        verbose_name='Решен',
        default=False,
    )

    class Meta:
        verbose_name = 'Тикет'
        verbose_name_plural = 'Тикеты'

    def __str__(self):
        return self.title


class Message(models.Model):
    class Sender(models.TextChoices):
        USER = 'user', 'Пользователь'
        SUPPORT_MANAGER = 'supp', 'Менеджер поддержки'

    id = models.BigAutoField(
        verbose_name='ID',
        primary_key=True,
        unique=True,
    )
    created_at = models.DateTimeField(
        verbose_name='Создано в',
        auto_now_add=True,
    )
    ticket = models.ForeignKey(
        verbose_name='Тикет',
        to=Ticket,
        on_delete=models.CASCADE,
        related_name='messages',
    )
    sender = models.CharField(
        verbose_name='Отправитель',
        max_length=4,
        choices=Sender.choices,
        default=Sender.USER,
    )
    text = models.CharField(
        verbose_name='Текст',
        max_length=4096,
    )
    viewed = models.BooleanField(
        verbose_name='Просмотрено',
        default=False,
    )

    class Meta:
        verbose_name = 'Сообщение'
        verbose_name_plural = 'Сообщения'

    def __str__(self):
        text = f'{self.text[:20]}...'
        sender = self.sender_instance
        if sender is None:
            return text
        return f'{sender}: {text}'

    @property
    def sender_instance(self):
        match self.sender:
            case self.Sender.USER:
                return self.ticket.user
            case self.Sender.SUPPORT_MANAGER:
                return self.ticket.support_manager
            case _:
                return None



from django.contrib.auth.mixins import LoginRequiredMixin
from django.views.generic import TemplateView
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework.generics import ListCreateAPIView, RetrieveUpdateAPIView
from rest_framework.pagination import LimitOffsetPagination

from . import filters, models, serializers


class Support(LoginRequiredMixin, TemplateView):
    template_name = 'support/support_manager.html'
    login_url = '/admin/login/'


class TicketList(ListCreateAPIView):
    queryset = models.Ticket.objects.all().order_by(
        'resolved',
        'created_at',
    )
    serializer_class = serializers.Ticket
    filter_backends = [
        DjangoFilterBackend,
    ]
    filterset_class = filters.Ticket


class Ticket(RetrieveUpdateAPIView):
    queryset = models.Ticket.objects.all()
    lookup_field = 'id'
    lookup_url_kwarg = 'ticket_id'
    serializer_class = serializers.Ticket


class MessageList(ListCreateAPIView):
    serializer_class = serializers.Message
    pagination_class = LimitOffsetPagination

    def get_queryset(self):
        return models.Message.objects.filter(
            ticket__id=self.kwargs['ticket_id']
        ).order_by('-created_at')

    def perform_create(self, serializer: serializer_class):
        serializer.save(
            ticket=models.Ticket.objects.get(id=self.kwargs['ticket_id']),
        )



from django.contrib import admin

from . import models


class MessageInline(admin.TabularInline):
    model = models.Message
    extra = 0


@admin.register(models.Ticket)
class TicketAdmin(admin.ModelAdmin):
    list_display = [
        'id',
        'created_at',
        'title',
        'support_manager',
        'resolved',
    ]
    inlines = [
        MessageInline,
    ]



import django.contrib.auth.models
from rest_framework import serializers

from . import models


class SupportManager(serializers.ModelSerializer):
    id = serializers.IntegerField(
        read_only=True,
        source='pk',
    )

    class Meta:
        model = django.contrib.auth.models.User
        read_only_fields = [
            'first_name',
            'last_name',
        ]
        fields = [
            'id',
            'first_name',
            'last_name',
        ]


class TelegramUser(serializers.ModelSerializer):
    class Meta:
        model = models.get_telegram_user_model()
        read_only_fields = [
            'telegram_id',
        ]
        fields = [
            'telegram_id',
        ]


class Ticket(serializers.ModelSerializer):
    user = TelegramUser(
        read_only=True,
    )
    user_id = serializers.PrimaryKeyRelatedField(
        queryset=models.get_telegram_user_model().objects.all(),
        write_only=True,
        source='user',
    )
    support_manager = SupportManager(
        read_only=True,
    )

    class Meta:
        model = models.Ticket
        fields = '__all__'
        read_only_fields = [
            'id',
            'created_at',
        ]


class Message(serializers.ModelSerializer):
    class Meta:
        model = models.Message
        fields = '__all__'
        read_only_fields = [
            'id',
            'created_at',
            'ticket',
        ]



from django.apps import AppConfig
from django.db.models.signals import post_save


class SupportConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'support'
    verbose_name = 'Поддержка'

    def ready(self) -> None:
        from . import models, signals

        post_save.connect(signals.post_save_ticket, models.Ticket)
        post_save.connect(signals.post_save_message, models.Message)



def get_asgi_application():
    import django.core.asgi
    from channels.auth import AuthMiddlewareStack
    from channels.routing import ProtocolTypeRouter, URLRouter

    from .routing import ws_urlpatterns

    return ProtocolTypeRouter({
        'http': django.core.asgi.get_asgi_application(),
        'websocket': AuthMiddlewareStack(URLRouter(ws_urlpatterns)),
    })


__all__ = [
    'get_asgi_application',
]

__version__ = '1.0.4'
